@@ -2,11 +2,14 @@ use std::{collections::HashMap, fmt, net::SocketAddr};
 
 use anyhow::{anyhow, bail, Context, Result};
 use futures::stream::StreamExt;
-use glib::{ObjectExt, ToValue};
-use gstreamer::{Element, prelude::{ElementExt, ElementExtManual, GObjectExtManualGst, GstBinExt, GstObjectExt, PadExt}};
-use gstreamer_rtp::{prelude::RTPHeaderExtensionExt, RTPHeaderExtension};
+use glib::{Cast, ObjectExt, ToValue};
+use gstreamer::{
+  prelude::{ElementExt, ElementExtManual, GObjectExtManualGst, GstBinExt, GstObjectExt, PadExt},
+  Bin, Element, GhostPad,
+};
 #[cfg(feature = "log-rtp")]
 use gstreamer_rtp::RTPBuffer;
+use gstreamer_rtp::{prelude::RTPHeaderExtensionExt, RTPHeaderExtension};
 use itertools::Itertools;
 use jitsi_xmpp_parsers::{
   jingle::{Action, Content, Description, Jingle, Transport},
@@ -20,8 +23,8 @@ use pem::Pem;
 use rand::random;
 use rcgen::{Certificate, CertificateParams, PKCS_ECDSA_P256_SHA256};
 use ring::digest::{digest, SHA256};
-use tokio::{net::lookup_host, runtime::Handle, sync::oneshot};
-use tracing::{debug, error, warn};
+use tokio::{net::lookup_host, runtime::Handle, sync::oneshot, task::JoinHandle};
+use tracing::{debug, error, info, warn};
 use uuid::Uuid;
 use xmpp_parsers::{
   hashes::Algo,
@@ -107,7 +110,7 @@ impl Codec {
     }
   }
 
-  fn make_depay_name(&self) -> &'static str {
+  fn depayloader_name(&self) -> &'static str {
     match self.name {
       CodecName::Opus => "rtpopusdepay",
       CodecName::H264 => "rtph264depay",
@@ -116,7 +119,16 @@ impl Codec {
     }
   }
 
-  fn make_pay_name(&self) -> &'static str {
+  fn decoder_name(&self) -> &'static str {
+    match self.name {
+      CodecName::Opus => "opusdec",
+      CodecName::H264 => "avdec_h264",
+      CodecName::Vp8 => "vp8dec",
+      CodecName::Vp9 => "vp9dec",
+    }
+  }
+
+  fn payloader_name(&self) -> &'static str {
     match self.name {
       CodecName::Opus => "rtpopuspay",
       CodecName::H264 => "rtph264pay",
@@ -137,11 +149,12 @@ pub(crate) struct JingleSession {
   pipeline: gstreamer::Pipeline,
   audio_sink_element: gstreamer::Element,
   video_sink_element: gstreamer::Element,
-  remote_ssrc_map: HashMap<u32, Source>,
+  pub(crate) remote_ssrc_map: HashMap<u32, Source>,
   _ice_agent: nice::Agent,
   pub(crate) accept_iq_id: Option<String>,
   pub(crate) colibri_url: Option<String>,
   pub(crate) colibri_channel: Option<ColibriChannel>,
+  pub(crate) stats_handler_task: Option<JoinHandle<()>>,
   pipeline_state_null_rx: oneshot::Receiver<()>,
 }
 
@@ -547,6 +560,8 @@ impl JingleSession {
     let rtpbin = gstreamer::ElementFactory::make("rtpbin", Some("rtpbin"))?;
     rtpbin.set_property_from_str("rtp-profile", "savpf");
     rtpbin.set_property("autoremove", true);
+    rtpbin.set_property("do-lost", true);
+    rtpbin.set_property("do-sync-event", true);
     pipeline.add(&rtpbin)?;
 
     let nicesrc = gstreamer::ElementFactory::make("nicesrc", None)?;
@@ -582,8 +597,7 @@ impl JingleSession {
         let f = || {
           debug!("rtpbin request-pt-map {:?}", values);
           let pt = values[2].get::<u32>()? as u8;
-          let mut caps = gstreamer::Caps::builder("application/x-rtp")
-            .field("payload", pt as i32);
+          let mut caps = gstreamer::Caps::builder("application/x-rtp").field("payload", pt as i32);
           for codec in codecs.iter() {
             if codec.is(pt) {
               if codec.is_audio() {
@@ -603,7 +617,8 @@ impl JingleSession {
                 caps = caps
                   .field("media", "video")
                   .field("clock-rate", 90000)
-                  .field("encoding-name", codec.encoding_name());
+                  .field("encoding-name", codec.encoding_name())
+                  .field("rtcp-fb-nack-pli", true);
                 if let Some(hdrext) = video_hdrext_transport_cc {
                   caps = caps.field(&format!("extmap-{}", hdrext), RTP_HDREXT_TRANSPORT_CC);
                 }
@@ -639,6 +654,7 @@ impl JingleSession {
 
     let handle = Handle::current();
     let jingle_session = conference.jingle_session.clone();
+    let buffer_size = conference.config.buffer_size;
     rtpbin.connect("new-jitterbuffer", false, move |values| {
       let handle = handle.clone();
       let jingle_session = jingle_session.clone();
@@ -668,6 +684,8 @@ impl JingleSession {
         if source.media_type == MediaType::Video && source.participant_id.is_some() {
           debug!("enabling RTX for ssrc {}", ssrc);
           rtpjitterbuffer.set_property("do-retransmission", true);
+          rtpjitterbuffer.set_property("drop-on-latency", true);
+          rtpjitterbuffer.set_property("latency", buffer_size);
         }
         Ok::<_, anyhow::Error>(())
       };
@@ -795,14 +813,14 @@ impl JingleSession {
 
             debug!("pad added for remote source: {:?}", source);
 
-            let source_element = match source.media_type {
+            let depayloader = match source.media_type {
               MediaType::Audio => {
                 let codec = codecs
                   .iter()
                   .filter(|codec| codec.is_audio())
                   .find(|codec| codec.is(pt));
                 if let Some(codec) = codec {
-                  gstreamer::ElementFactory::make(codec.make_depay_name(), None)?
+                  gstreamer::ElementFactory::make(codec.depayloader_name(), None)?
                 }
                 else {
                   bail!("received audio with unsupported PT {}", pt);
@@ -814,7 +832,7 @@ impl JingleSession {
                   .filter(|codec| codec.is_video())
                   .find(|codec| codec.is(pt));
                 if let Some(codec) = codec {
-                  gstreamer::ElementFactory::make(codec.make_depay_name(), None)?
+                  gstreamer::ElementFactory::make(codec.depayloader_name(), None)?
                 }
                 else {
                   bail!("received video with unsupported PT {}", pt);
@@ -822,8 +840,8 @@ impl JingleSession {
               },
             };
 
-            source_element.set_property("auto-header-extension", false);
-            source_element.connect("request-extension", false, move |values| {
+            depayloader.set_property("auto-header-extension", false);
+            depayloader.connect("request-extension", false, move |values| {
               let f = || {
                 let ext_id: u32 = values[1].get()?;
                 let ext_uri: String = values[2].get()?;
@@ -843,25 +861,152 @@ impl JingleSession {
             });
 
             pipeline
-              .add(&source_element)
+              .add(&depayloader)
               .context("failed to add depayloader to pipeline")?;
-            source_element.sync_state_with_parent()?;
+            depayloader.sync_state_with_parent()?;
             debug!("created depayloader");
             rtpbin
-              .link_pads(Some(&pad_name), &source_element, None)
+              .link_pads(Some(&pad_name), &depayloader, None)
               .context(format!("failed to link rtpbin.{} to depayloader", pad_name))?;
 
             debug!("linked rtpbin.{} to depayloader", pad_name);
 
             debug!("rtpbin pads:\n{}", dump_pads(&rtpbin));
 
-            let src_pad = source_element
-              .static_pad("src")
-              .context("depayloader has no src pad")?;
+            let queue = gstreamer::ElementFactory::make("queue", None)?;
+            pipeline
+              .add(&queue)
+              .context("failed to add queue to pipeline")?;
+            queue.sync_state_with_parent()?;
+            depayloader
+              .link(&queue)
+              .context("failed to link depayloader to queue")?;
+
+            let decoder = match source.media_type {
+              MediaType::Audio => {
+                let codec = codecs
+                  .iter()
+                  .filter(|codec| codec.is_audio())
+                  .find(|codec| codec.is(pt));
+                if let Some(codec) = codec {
+                  gstreamer::ElementFactory::make(codec.decoder_name(), None)?
+                  // TODO: fec
+                }
+                else {
+                  bail!("received audio with unsupported PT {}", pt);
+                }
+              },
+              MediaType::Video => {
+                let codec = codecs
+                  .iter()
+                  .filter(|codec| codec.is_video())
+                  .find(|codec| codec.is(pt));
+                if let Some(codec) = codec {
+                  let decoder = gstreamer::ElementFactory::make(codec.decoder_name(), None)?;
+                  decoder.set_property("automatic-request-sync-points", true);
+                  decoder.set_property_from_str(
+                    "automatic-request-sync-point-flags",
+                    "GST_VIDEO_DECODER_REQUEST_SYNC_POINT_CORRUPT_OUTPUT",
+                  );
+                  decoder
+                }
+                else {
+                  bail!("received video with unsupported PT {}", pt);
+                }
+              },
+            };
+
+            pipeline
+              .add(&decoder)
+              .context("failed to add decoder to pipeline")?;
+            decoder.sync_state_with_parent()?;
+            queue
+              .link(&decoder)
+              .context("failed to link queue to decoder")?;
+
+            let src_pad = match source.media_type {
+              MediaType::Audio => decoder
+                .static_pad("src")
+                .context("decoder has no src pad")?,
+              MediaType::Video => {
+                let videoscale = gstreamer::ElementFactory::make("videoscale", None)?;
+                pipeline
+                  .add(&videoscale)
+                  .context("failed to add videoscale to pipeline")?;
+                videoscale.sync_state_with_parent()?;
+                decoder
+                  .link(&videoscale)
+                  .context("failed to link decoder to videoscale")?;
+
+                let capsfilter = gstreamer::ElementFactory::make("capsfilter", None)?;
+                capsfilter.set_property_from_str(
+                  "caps",
+                  &format!(
+                    "video/x-raw, width={}, height={}",
+                    conference.config.recv_video_scale_width,
+                    conference.config.recv_video_scale_height
+                  ),
+                );
+                pipeline
+                  .add(&capsfilter)
+                  .context("failed to add capsfilter to pipeline")?;
+                capsfilter.sync_state_with_parent()?;
+                videoscale
+                  .link(&capsfilter)
+                  .context("failed to link videoscale to capsfilter")?;
+
+                let videoconvert = gstreamer::ElementFactory::make("videoconvert", None)?;
+                pipeline
+                  .add(&videoconvert)
+                  .context("failed to add videoconvert to pipeline")?;
+                videoconvert.sync_state_with_parent()?;
+                capsfilter
+                  .link(&videoconvert)
+                  .context("failed to link capsfilter to videoconvert")?;
+
+                videoconvert
+                  .static_pad("src")
+                  .context("videoconvert has no src pad")?
+              },
+            };
 
             if let Some(participant_id) = source.participant_id {
               handle.block_on(conference.ensure_participant(&participant_id))?;
-              if let Some(participant_bin) =
+              let maybe_sink_element = match source.media_type {
+                MediaType::Audio => {
+                  handle.block_on(conference.remote_participant_audio_sink_element())
+                },
+                MediaType::Video => {
+                  handle.block_on(conference.remote_participant_video_sink_element())
+                },
+              };
+              if let Some(sink_element) = maybe_sink_element {
+                let sink_pad = sink_element
+                  .request_pad_simple("sink_%u")
+                  .context("no suitable sink pad provided by sink element in recv pipeline")?;
+                let ghost_pad = GhostPad::with_target(
+                  Some(&format!(
+                    "participant_{}_{:?}",
+                    participant_id, source.media_type
+                  )),
+                  &sink_pad,
+                )?;
+                let bin: Bin = sink_element
+                  .parent()
+                  .context("sink element has no parent")?
+                  .downcast()
+                  .map_err(|_| anyhow!("sink element's parent is not a bin"))?;
+                bin.add_pad(&ghost_pad)?;
+
+                src_pad
+                  .link(&ghost_pad)
+                  .context("failed to link decode chain to participant bin from recv pipeline")?;
+                info!(
+                  "linked {}/{:?} to new pad in recv pipeline",
+                  participant_id, source.media_type
+                );
+              }
+              else if let Some(participant_bin) =
                 pipeline.by_name(&format!("participant_{}", participant_id))
               {
                 let sink_pad_name = match source.media_type {
@@ -869,18 +1014,23 @@ impl JingleSession {
                   MediaType::Video => "video",
                 };
                 if let Some(sink_pad) = participant_bin.static_pad(sink_pad_name) {
-                  debug!("linking depayloader to participant bin");
-                  src_pad.link(&sink_pad)?;
+                  src_pad.link(&sink_pad).context(
+                    "failed to link decode chain to participant bin from recv participant pipeline",
+                  )?;
+                  info!(
+                    "linked {}/{:?} to recv participant pipeline",
+                    participant_id, source.media_type
+                  );
                 }
                 else {
                   warn!(
-                    "no {} sink pad in {} participant bin",
+                    "no {} sink pad on {} participant bin in recv participant pipeline",
                     sink_pad_name, participant_id
                   );
                 }
               }
               else {
-                debug!("no participant bin for {}", participant_id);
+                warn!("no pipeline handled new participant: {}", participant_id);
               }
             }
             else {
@@ -888,11 +1038,14 @@ impl JingleSession {
             }
 
             if !src_pad.is_linked() {
-              debug!("nothing linked to depayloader, adding fakesink");
+              debug!("nothing linked to decoder, adding fakesink");
               let fakesink = gstreamer::ElementFactory::make("fakesink", None)?;
               pipeline.add(&fakesink)?;
               fakesink.sync_state_with_parent()?;
-              source_element.link(&fakesink)?;
+              let sink_pad = fakesink
+                .static_pad("sink")
+                .context("fakesink has no sink pad")?;
+              src_pad.link(&sink_pad)?;
             }
 
             gstreamer::debug_bin_to_dot_file(
@@ -916,7 +1069,7 @@ impl JingleSession {
 
     let opus = codecs.iter().find(|codec| codec.name == CodecName::Opus);
     let audio_sink_element = if let Some(opus) = opus {
-      let audio_sink_element = gstreamer::ElementFactory::make(opus.make_pay_name(), None)?;
+      let audio_sink_element = gstreamer::ElementFactory::make(opus.payloader_name(), None)?;
       audio_sink_element.set_property("pt", opus.pt as u32);
       audio_sink_element
     }
@@ -957,7 +1110,7 @@ impl JingleSession {
     let codec_name = conference.config.video_codec.as_str();
     let codec = codecs.iter().find(|codec| codec.is_codec(codec_name));
     let video_sink_element = if let Some(codec) = codec {
-      let element = gstreamer::ElementFactory::make(codec.make_pay_name(), None)?;
+      let element = gstreamer::ElementFactory::make(codec.payloader_name(), None)?;
       element.set_property("pt", codec.pt as u32);
       if codec.name == CodecName::H264 {
         element.set_property_from_str("aggregate-mode", "zero-latency");
@@ -1022,62 +1175,66 @@ impl JingleSession {
     pipeline.add(&rtcp_send_identity)?;
 
     #[cfg(feature = "log-rtp")]
-    if conference.config.log_rtp {
-      debug!("setting up RTP/RTCP packet logging");
+    {
+      if conference.config.log_rtp {
+        debug!("setting up RTP packet logging");
 
-      let make_rtp_logger = |direction: &'static str| {
-        move |values: &[glib::Value]| -> Option<glib::Value> {
-          let f = || {
-            let buffer: gstreamer::Buffer = values[1].get()?;
-            let rtp_buffer = RTPBuffer::from_buffer_readable(&buffer)?;
-            debug!(
-              ssrc=rtp_buffer.ssrc(),
-              pt=rtp_buffer.payload_type(),
-              seq=rtp_buffer.seq(),
-              ts=rtp_buffer.timestamp(),
-              marker=rtp_buffer.is_marker(),
-              extension=rtp_buffer.is_extension(),
-              payload_size=rtp_buffer.payload_size(),
-              "RTP {}",
-              direction,
-            );
-            Ok::<_, anyhow::Error>(())
-          };
-          if let Err(e) = f() {
-            warn!("RTP {}: {:?}", direction, e);
+        let make_rtp_logger = |direction: &'static str| {
+          move |values: &[glib::Value]| -> Option<glib::Value> {
+            let f = || {
+              let buffer: gstreamer::Buffer = values[1].get()?;
+              let rtp_buffer = RTPBuffer::from_buffer_readable(&buffer)?;
+              debug!(
+                ssrc = rtp_buffer.ssrc(),
+                pt = rtp_buffer.payload_type(),
+                seq = rtp_buffer.seq(),
+                ts = rtp_buffer.timestamp(),
+                marker = rtp_buffer.is_marker(),
+                extension = rtp_buffer.is_extension(),
+                payload_size = rtp_buffer.payload_size(),
+                "RTP {}",
+                direction,
+              );
+              Ok::<_, anyhow::Error>(())
+            };
+            if let Err(e) = f() {
+              warn!("RTP {}: {:?}", direction, e);
+            }
+            None
           }
-          None
-        }
-      };
+        };
 
-      let make_rtcp_logger = |direction: &'static str| {
-        move |values: &[glib::Value]| -> Option<glib::Value> {
-          let f = || {
-            let buffer: gstreamer::Buffer = values[1].get()?;
-            let mut buf = [0u8; 1500];
-            buffer.copy_to_slice(0, &mut buf[..buffer.size()]).map_err(|_| anyhow!("invalid RTCP packet size"))?;
-            let decoded = rtcp::packet::unmarshal(&mut &buf[..buffer.size()])?;
-            debug!(
-              "RTCP {} size={}\n{:#?}",
-              direction,
-              buffer.size(),
-              decoded,
-            );
-            Ok::<_, anyhow::Error>(())
-          };
-          if let Err(e) = f() {
-            warn!("RTCP {}: {:?}", direction, e);
+        rtp_recv_identity.connect("handoff", false, make_rtp_logger("RECV"));
+        rtp_send_identity.connect("handoff", false, make_rtp_logger("SEND"));
+      }
+
+      if conference.config.log_rtcp {
+        debug!("setting up RTCP packet logging");
+
+        let make_rtcp_logger = |direction: &'static str| {
+          move |values: &[glib::Value]| -> Option<glib::Value> {
+            let f = || {
+              let buffer: gstreamer::Buffer = values[1].get()?;
+              let mut buf = [0u8; 1500];
+              buffer
+                .copy_to_slice(0, &mut buf[..buffer.size()])
+                .map_err(|_| anyhow!("invalid RTCP packet size"))?;
+              let decoded = rtcp::packet::unmarshal(&mut &buf[..buffer.size()])?;
+              debug!("RTCP {} size={}\n{:#?}", direction, buffer.size(), decoded,);
+              Ok::<_, anyhow::Error>(())
+            };
+            if let Err(e) = f() {
+              warn!("RTCP {}: {:?}", direction, e);
+            }
+            None
           }
-          None
-        }
-      };
+        };
 
-      rtp_recv_identity.connect("handoff", false, make_rtp_logger("RECV"));
-      rtp_send_identity.connect("handoff", false, make_rtp_logger("SEND"));
-      rtcp_recv_identity.connect("handoff", false, make_rtcp_logger("RECV"));
-      rtcp_send_identity.connect("handoff", false, make_rtcp_logger("SEND"));
+        rtcp_recv_identity.connect("handoff", false, make_rtcp_logger("RECV"));
+        rtcp_send_identity.connect("handoff", false, make_rtcp_logger("SEND"));
+      }
     }
-    
+
     debug!("link dtlssrtpdec -> rtpbin");
     dtlssrtpdec.link_pads(Some("rtp_src"), &rtp_recv_identity, None)?;
     rtp_recv_identity.link_pads(None, &rtpbin, Some("recv_rtp_sink_0"))?;
@@ -1320,6 +1477,7 @@ impl JingleSession {
       accept_iq_id: Some(accept_iq_id),
       colibri_url: ice_transport.web_socket.clone().map(|ws| ws.url),
       colibri_channel: None,
+      stats_handler_task: None,
       pipeline_state_null_rx,
     })
   }
@@ -1371,12 +1529,21 @@ fn dump_pads(element: &Element) -> String {
   element
     .pads()
     .into_iter()
-    .map(|pad| format!(
-      "  {}, peer={}.{}, caps=\"{}\"",
-      pad.name(),
-      pad.peer().and_then(|peer| peer.parent_element()).map(|element| element.name().to_string()).unwrap_or_default(),
-      pad.peer().map(|peer| peer.name().to_string()).unwrap_or_default(),
-      pad.caps().map(|caps| caps.to_string()).unwrap_or_default(),
-    ))
+    .map(|pad| {
+      format!(
+        "  {}, peer={}.{}, caps=\"{}\"",
+        pad.name(),
+        pad
+          .peer()
+          .and_then(|peer| peer.parent_element())
+          .map(|element| element.name().to_string())
+          .unwrap_or_default(),
+        pad
+          .peer()
+          .map(|peer| peer.name().to_string())
+          .unwrap_or_default(),
+        pad.caps().map(|caps| caps.to_string()).unwrap_or_default(),
+      )
+    })
     .join("\n")
-}
\ No newline at end of file
+}
